@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2014 Canonical
+ * Copyright (C) 2011-2015 Canonical
  * Hugely modified parts from powertop-1.13, Copyright 2007, Intel Corporation
  *
  * This program is free software; you can redistribute it and/or
@@ -16,12 +16,15 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
+ * eventstat by Colin Ian King <colin.king@canonical.com>
  */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdbool.h>
 #include <string.h>
+#include <stdint.h>
+#include <inttypes.h>
 #include <signal.h>
 #include <unistd.h>
 #include <time.h>
@@ -33,9 +36,8 @@
 #include <fcntl.h>
 #include <libgen.h>
 #include <math.h>
+#include <float.h>
 
-#define APP_NAME		"eventstat"
-#define TIMER_STATS		"/proc/timer_stats"
 #define TABLE_SIZE		(32771)		/* Should be a prime */
 
 #define OPT_QUIET		(0x00000001)
@@ -71,26 +73,30 @@ typedef struct timer_info {
 	char		*callback;	/* Kernel timer callback func */
 	char		*ident;		/* Unique identity */
 	bool		kernel_thread;	/* True if task is a kernel thread */
-	unsigned long	total;		/* Total number of events */
+	uint64_t	total;		/* Total number of events */
+	double		time_total;	/* Total time */
 } timer_info_t;
 
 typedef struct timer_stat {
-	unsigned long	count;		/* Number of events */
-	unsigned long	delta;		/* Change in events since last time */
+	uint64_t	count;		/* Number of events */
+	int64_t		delta;		/* Change in events since last time */
+	double		time;		/* Time of sample */
+	double		time_delta;	/* Change in time since last time */
 	timer_info_t	*info;		/* Timer info */
 	struct timer_stat *next;	/* Next timer stat in hash table */
-	struct timer_stat *sorted_freq_next;	/* Next timer stat in event frequency sorted list */
+	struct timer_stat *sorted_freq_next; /* Next timer stat in event frequency sorted list */
 } timer_stat_t;
 
 /* sample delta item as an element of the sample_delta_list_t */
 typedef struct sample_delta_item {
-	unsigned long	delta;		/* difference in timer events between old and new */
+	int64_t		delta;		/* difference in timer events between old and new */
+	double		time_delta;	/* difference in time between old and new */
 	timer_info_t	*info;		/* timer this refers to */
 } sample_delta_item_t;
 
 /* list of sample_delta_items */
 typedef struct sample_delta_list {
-	struct timeval 	whence;		/* when the sample was taken */
+	double		whence;		/* when the sample was taken */
 	list_t		list;
 } sample_delta_list_t;
 
@@ -101,14 +107,171 @@ typedef struct {
 
 #define KERN_TASK_INFO(str)		{ str, sizeof(str) - 1 }
 
+static const char *app_name = "eventstat";
+static const char *proc_timer_stats = "/proc/timer_stats";
+
 static list_t timer_info_list;		/* cache list of timer_info */
 static list_t sample_list;		/* list of samples, sorted in sample time order */
 static char *csv_results;		/* results in comma separated values */
 static volatile bool stop_eventstat = false;	/* set by sighandler */
 static double  opt_threshold;		/* ignore samples with event delta less than this */
-static unsigned int opt_flags;		/* option flags */
+static uint32_t opt_flags;		/* option flags */
 static bool sane_procs;			/* false if we are in a container */
 
+/*
+ *  Attempt to catch a range of signals so
+ *  we can clean
+ */
+static const int signals[] = {
+	/* POSIX.1-1990 */
+#ifdef SIGHUP
+	SIGHUP,
+#endif
+#ifdef SIGINT
+	SIGINT,
+#endif
+#ifdef SIGQUIT
+	SIGQUIT,
+#endif
+#ifdef SIGILL
+	SIGILL,
+#endif
+#ifdef SIGABRT
+	SIGABRT,
+#endif
+#ifdef SIGFPE
+	SIGFPE,
+#endif
+#ifdef SIGSEGV
+	SIGSEGV,
+#endif
+#ifdef SIGTERM
+	SIGTERM,
+#endif
+#ifdef SIGUSR1
+	SIGUSR1,
+#endif
+#ifdef SIGUSR2
+	SIGUSR2,
+	/* POSIX.1-2001 */
+#endif
+#ifdef SIGBUS
+	SIGBUS,
+#endif
+#ifdef SIGXCPU
+	SIGXCPU,
+#endif
+#ifdef SIGXFSZ
+	SIGXFSZ,
+#endif
+	/* Linux various */
+#ifdef SIGIOT
+	SIGIOT,
+#endif
+#ifdef SIGSTKFLT
+	SIGSTKFLT,
+#endif
+#ifdef SIGPWR
+	SIGPWR,
+#endif
+#ifdef SIGINFO
+	SIGINFO,
+#endif
+#ifdef SIGVTALRM
+	SIGVTALRM,
+#endif
+	-1,
+};
+
+/*
+ *  set_timer_stat()
+ *	enable/disable timer stat
+ */
+static void set_timer_stat(const char *str, const bool carp)
+{
+	int fd;
+	ssize_t len = (ssize_t)strlen(str);
+
+	if ((fd = open(proc_timer_stats, O_WRONLY, S_IRUSR | S_IWUSR)) < 0) {
+		if (carp) {
+			fprintf(stderr, "Cannot open %s, errno=%d (%s)\n",
+				proc_timer_stats, errno, strerror(errno));
+			exit(EXIT_FAILURE);
+		} else {
+			return;
+		}
+	}
+	if (write(fd, str, len) != len) {
+		close(fd);
+		if (carp) {
+			fprintf(stderr, "Cannot write to %s, errno=%d (%s)\n",
+				proc_timer_stats, errno, strerror(errno));
+			exit(EXIT_FAILURE);
+		} else {
+			return;
+		}
+	}
+	(void)close(fd);
+}
+
+
+/*
+ *  Stop gcc complaining about no return func
+ */
+static void eventstat_exit(const int status) __attribute__ ((noreturn));
+
+/*
+ *  eventstat_exit()
+ *	exit and set timer stat to 0
+ */
+static void eventstat_exit(const int status)
+{
+	set_timer_stat("0\n", false);
+
+	exit(status);
+}
+
+
+/*
+ *  timeval_to_double
+ *	timeval to a double (in seconds)
+ */
+static inline double timeval_to_double(const struct timeval *const tv)
+{
+	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
+}
+
+
+/*
+ *  double_to_timeval
+ *	seconds in double to timeval
+ */
+static inline struct timeval double_to_timeval(const double val)
+{
+	struct timeval tv;
+
+	tv.tv_sec = val;
+	tv.tv_usec = (val - (time_t)val) * 1000000.0;
+
+	return tv;
+}
+
+/*
+ *  gettime_to_double()
+ *      get time as a double
+ */
+static double gettime_to_double(void)
+{
+	struct timeval tv;
+
+	if (gettimeofday(&tv, NULL) < 0) {
+		fprintf(stderr, "gettimeofday failed: errno=%d (%s)\n",
+			errno, strerror(errno));
+		eventstat_exit(EXIT_FAILURE);
+	}
+	return timeval_to_double(&tv);
+}
+
 /*
  *  sane_proc_pid_info()
  *	detect if proc info mapping from /proc/timer_stats
@@ -145,103 +308,11 @@ static bool sane_proc_pid_info(void)
 		}
 	}
 
-	fclose(fp);
+	(void)fclose(fp);
 
 	return ret;
 }
 
-/*
- *  timeval_sub()
- *	timeval a - b
- */
-static struct timeval timeval_sub(const struct timeval *a, const struct timeval *b)
-{
-	struct timeval ret, _b;
-
-	_b.tv_sec = b->tv_sec;
-	_b.tv_usec = b->tv_usec;
-
-	if (a->tv_usec < _b.tv_usec) {
-		int nsec = ((_b.tv_usec - a->tv_usec) / 1000000) + 1;
-		_b.tv_sec += nsec;
-		_b.tv_usec -= (1000000 * nsec);
-	}
-	if (a->tv_usec - _b.tv_usec > 1000000) {
-		int nsec = (a->tv_usec - _b.tv_usec) / 1000000;
-		_b.tv_sec -= nsec;
-		_b.tv_usec += (1000000 * nsec);
-	}
-
-	ret.tv_sec = a->tv_sec - _b.tv_sec;
-	ret.tv_usec = a->tv_usec - _b.tv_usec;
-
-	return ret;
-}
-
-/*
- *  timeval_add()
- *	timeval a + b
- */
-static struct timeval timeval_add(const struct timeval *a, const struct timeval *b)
-{
-	struct timeval ret;
-
-	ret.tv_sec = a->tv_sec + b->tv_sec;
-	ret.tv_usec = a->tv_usec + b->tv_usec;
-	if (ret.tv_usec > 1000000) {
-		int nsec = (ret.tv_usec / 1000000);
-		ret.tv_sec += nsec;
-		ret.tv_usec -= (1000000 * nsec);
-	}
-
-	return ret;
-}
-
-/*
- *  timeval_double
- *	timeval to a double
- */
-static inline double timeval_double(const struct timeval *tv)
-{
-	return (double)tv->tv_sec + ((double)tv->tv_usec / 1000000.0);
-}
-
-/*
- *  set_timer_stat()
- *	enable/disable timer stat
- */
-static void set_timer_stat(const char *str, const bool carp)
-{
-	FILE *fp;
-
-	if ((fp = fopen(TIMER_STATS, "w")) == NULL) {
-		if (carp) {
-			fprintf(stderr, "Cannot write to %s\n",TIMER_STATS);
-			exit(EXIT_FAILURE);
-		} else {
-			return;
-		}
-	}
-	fprintf(fp, "%s\n", str);
-	fclose(fp);
-}
-
-/*
- *  Stop gcc complaining about no return func
- */
-static void eventstat_exit(const int status) __attribute__ ((noreturn));
-
-/*
- *  eventstat_exit()
- *	exit and set timer stat to 0
- */
-static void eventstat_exit(const int status)
-{
-	set_timer_stat("0", false);
-
-	exit(status);
-}
-
 /*
  *  list_init()
  *	initialize list
@@ -298,14 +369,15 @@ static void list_free(list_t *list, const list_link_free_t freefunc)
 }
 
 /*
- *  handle_sigint()
- *      catch SIGINT and flag a stop
+ *  handle_sig()
+ *      catch signal, flag a stop and restore timer stat
  */
-static void handle_sigint(int dummy)
+static void handle_sig(int dummy)
 {
 	(void)dummy;	/* Stop unused parameter warning with -Wextra */
 
 	stop_eventstat = true;
+	set_timer_stat("0\n", false);
 }
 
 /*
@@ -333,7 +405,9 @@ static void samples_free(void)
  *  sample_add()
  *	add a timer_stat's delta and info field to a list at time position whence
  */
-static void sample_add(timer_stat_t *timer_stat, struct timeval *whence)
+static void sample_add(
+	timer_stat_t *timer_stat,
+	const double whence)
 {
 	link_t	*link;
 	bool	found = false;
@@ -345,8 +419,7 @@ static void sample_add(timer_stat_t *timer_stat, struct timeval *whence)
 
 	for (link = sample_list.head; link; link = link->next) {
 		sdl = (sample_delta_list_t*)link->data;
-		if ((sdl->whence.tv_sec == whence->tv_sec) &&
-		    (sdl->whence.tv_usec == whence->tv_usec)) {
+		if (sdl->whence == whence) {
 			found = true;
 			break;
 		}
@@ -361,7 +434,7 @@ static void sample_add(timer_stat_t *timer_stat, struct timeval *whence)
 			fprintf(stderr, "Cannot allocate sample delta list\n");
 			eventstat_exit(EXIT_FAILURE);
 		}
-		sdl->whence = *whence;
+		sdl->whence = whence;
 		list_append(&sample_list, sdl);
 	}
 
@@ -371,6 +444,7 @@ static void sample_add(timer_stat_t *timer_stat, struct timeval *whence)
 		eventstat_exit(EXIT_FAILURE);
 	}
 	sdi->delta = timer_stat->delta;
+	sdi->time_delta = timer_stat->time_delta;
 	sdi->info  = timer_stat->info;
 
 	list_append(&sdl->list, sdi);
@@ -401,7 +475,10 @@ static int info_compare_total(const void *item1, const void *item2)
 	timer_info_t **info1 = (timer_info_t **)item1;
 	timer_info_t **info2 = (timer_info_t **)item2;
 
-	return (*info2)->total - (*info1)->total;
+	if ((*info2)->total == (*info1)->total)
+		return 0;
+
+	return ((*info2)->total > (*info1)->total) ? 1 : -1;
 }
 
 /*
@@ -479,10 +556,10 @@ static char *get_pid_cmdline(const pid_t id)
 		return NULL;
 
 	if ((ret = read(fd, buffer, sizeof(buffer))) <= 0) {
-		close(fd);
+		(void)close(fd);
 		return NULL;
 	}
-	close(fd);
+	(void)close(fd);
 
 	buffer[sizeof(buffer)-1] = '\0';
 
@@ -512,11 +589,16 @@ static char *get_pid_cmdline(const pid_t id)
 	return strdup(buffer);
 }
 
+static inline double duration_round(const double duration)
+{
+	return floor((duration * 100.0) + 0.5) / 100.0;
+}
+
 /*
  *  samples_dump()
  *	dump out collected sample information
  */
-static void samples_dump(const char *filename, const struct timeval *duration)
+static void samples_dump(const char *filename)
 {
 	sample_delta_list_t	*sdl;
 	timer_info_t **sorted_timer_infos;
@@ -524,9 +606,9 @@ static void samples_dump(const char *filename, const struct timeval *duration)
 	size_t i = 0;
 	size_t n = timer_info_list.length;
 	FILE *fp;
-	unsigned long count = 0;
-	double dur;
-	bool dur_zero;
+	uint64_t count = 0;
+	double first_time = -1.0;
+	double duration;
 
 	if (filename == NULL)
 		return;
@@ -575,28 +657,31 @@ static void samples_dump(const char *filename, const struct timeval *duration)
 
 	fprintf(fp, "Total:");
 	for (i = 0; i < n; i++)
-		fprintf(fp, ",%lu", sorted_timer_infos[i]->total);
+		fprintf(fp, ",%" PRIu64, sorted_timer_infos[i]->total);
 	fprintf(fp, "\n");
 
-	/*
-	 *  duration - if -C option is used then don't scale by the
-	 *  per sample duration time, instead give the raw sample count
-	 *  by scaling by 1.0 (i.e. no scaling).
-	 */
-	dur = (opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : timeval_double(duration);
-	dur_zero = (duration->tv_sec == 0) && (duration->tv_usec == 0);
-
 	for (link = sample_list.head; link; link = link->next) {
 		count++;
 		sdl = (sample_delta_list_t*)link->data;
-		fprintf(fp, "%f", timeval_double(&sdl->whence));
+
+		if (first_time < 0)
+			first_time = sdl->whence;
+		fprintf(fp, "%f", duration_round(sdl->whence - first_time));
 
 		/* Scan in timer info order to be consistent for all sdl rows */
 		for (i = 0; i < n; i++) {
 			sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
-			if (sdi)
-				fprintf(fp, ",%f", dur_zero ? 0.0 : (double)sdi->delta / dur);
-			else
+
+			/*
+			 *  duration - if -C option is used then don't scale by the
+			 *  per sample duration time, instead give the raw sample count
+			 *  by scaling by 1.0 (i.e. no scaling).
+			 */
+			if (sdi) {
+				duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
+				fprintf(fp, ",%f", duration == 0.0 ? 0.0 :
+					(double)sdi->delta / duration);
+			} else
 				fprintf(fp, ",%f", 0.0);
 		}
 		fprintf(fp, "\n");
@@ -608,35 +693,46 @@ static void samples_dump(const char *filename, const struct timeval *duration)
 	if (opt_flags & OPT_RESULT_STATS) {
 		fprintf(fp, "Min:");
 		for (i = 0; i < n; i++) {
-			unsigned long min = ~0;
+			double min = DBL_MAX;
 
 			for (link = sample_list.head; link; link = link->next) {
 				sdl = (sample_delta_list_t*)link->data;
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
-				if (sdi && min > sdi->delta)
-					min = sdi->delta;
+
+				if (sdi) {
+					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
+					double val = (duration == 0.0) ? 0.0: sdi->delta / duration;
+					if (min > val)
+						min = val;
+				}
 			}
-			fprintf(fp, ",%f", dur_zero ? 0.0 : (double)min / dur);
+			fprintf(fp, ",%f", min);
 		}
 		fprintf(fp, "\n");
 
 		fprintf(fp, "Max:");
 		for (i = 0; i < n; i++) {
-			unsigned long max = 0;
+			double max = DBL_MIN;
 
 			for (link = sample_list.head; link; link = link->next) {
 				sdl = (sample_delta_list_t*)link->data;
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
-				if (sdi && max < sdi->delta)
-					max = sdi->delta;
+
+				if (sdi) {
+					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
+					double val = (duration == 0.0) ? 0.0: sdi->delta / duration;
+					if (max < val)
+						max = val;
+				}
 			}
-			fprintf(fp, ",%f", dur_zero ? 0.0 : (double)max / dur);
+			fprintf(fp, ",%f", max);
 		}
 		fprintf(fp, "\n");
 
 		fprintf(fp, "Average:");
 		for (i = 0; i < n; i++)
-			fprintf(fp, ",%f", dur_zero ? 0.0 : ((double)sorted_timer_infos[i]->total / dur) / (double)count);
+			fprintf(fp, ",%f", count == 0 ? 0.0 :
+				(double)sorted_timer_infos[i]->total / count);
 		fprintf(fp, "\n");
 
 		/*
@@ -651,7 +747,9 @@ static void samples_dump(const char *filename, const struct timeval *duration)
 				sdl = (sample_delta_list_t*)link->data;
 				sample_delta_item_t *sdi = sample_find(sdl, sorted_timer_infos[i]);
 				if (sdi) {
-					double diff = dur_zero ? 0.0 : ((double)sdi->delta - average) / dur;
+					double duration = duration_round((opt_flags & OPT_SAMPLE_COUNT) ? 1.0 : sdi->time_delta);
+					double diff = duration == 0.0 ? 0.0 :
+						((double)sdi->delta - average) / duration;
 					diff = diff * diff;
 					sum += diff;
 				}
@@ -663,7 +761,7 @@ static void samples_dump(const char *filename, const struct timeval *duration)
 	}
 
 	free(sorted_timer_infos);
-	fclose(fp);
+	(void)fclose(fp);
 }
 
 /*
@@ -697,6 +795,7 @@ static timer_info_t *timer_info_find(const timer_info_t *new_info)
 	info->ident = strdup(new_info->ident);
 	info->kernel_thread = new_info->kernel_thread;
 	info->total = new_info->total;
+	info->time_total = new_info->time_total;
 
 	if (info->task == NULL ||
 	    info->func == NULL ||
@@ -742,14 +841,14 @@ static void timer_info_list_free(void)
  *  hash_pjw()
  *	Hash a string, from Aho, Sethi, Ullman, Compiling Techniques.
  */
-static unsigned long hash_pjw(const char *str)
+static uint32_t hash_pjw(const char *str)
 {
-  	unsigned long h = 0;
+  	uint32_t h = 0;
 
 	while (*str) {
-		unsigned long g;
+		uint32_t g;
 		h = (h << 4) + (*str);
-		if (0 != (g = h&0xf0000000)) {
+		if (0 != (g = h & 0xf0000000)) {
 			h = h ^ (g >> 24);
 			h = h ^ g;
 		}
@@ -788,7 +887,8 @@ static void timer_stat_free_contents(
  */
 static void timer_stat_add(
 	timer_stat_t *timer_stats[],	/* timer stat hash table */
-	const unsigned long count,	/* event count */
+	const double time_now,		/* time sample was taken */
+	const uint64_t count,		/* event count */
 	const pid_t pid,		/* PID of task */
 	char *task,			/* Name of task */
 	char *func,			/* Kernel function */
@@ -799,7 +899,7 @@ static void timer_stat_add(
 	timer_stat_t *ts;
 	timer_stat_t *ts_new;
 	timer_info_t info;
-	unsigned long h;
+	uint32_t h;
 
 	snprintf(buf, sizeof(buf), "%d:%s:%s:%s", pid, task, func, callback);
 	h = hash_pjw(buf);
@@ -825,10 +925,12 @@ static void timer_stat_add(
 	info.ident = buf;
 	info.kernel_thread = kernel_thread;
 	info.total = count;
+	info.time_total = 0.0;
 
-	ts_new->count  = count;
+	ts_new->count = count;
 	ts_new->info = timer_info_find(&info);
-	ts_new->next  = timer_stats[h];
+	ts_new->next = timer_stats[h];
+	ts_new->time = time_now;
 	ts_new->sorted_freq_next = NULL;
 
 	timer_stats[h] = ts_new;
@@ -889,15 +991,13 @@ static void timer_stat_sort_freq_add(
  *	silently die
  */
 static void timer_stat_diff(
-	struct timeval *duration,	/* time between each sample */
-	const long int n_lines,		/* number of lines to output */
-	struct timeval *whence,		/* nth sample */
+	const double duration,		/* time between each sample */
+	const int32_t n_lines,		/* number of lines to output */
+	const double whence,		/* nth sample */
 	timer_stat_t *timer_stats_old[],/* old timer stats samples */
 	timer_stat_t *timer_stats_new[])/* new timer stats samples */
 {
 	int i;
-	double dur = timeval_double(duration);
-
 	timer_stat_t *sorted = NULL;
 
 	for (i = 0; i < TABLE_SIZE; i++) {
@@ -908,13 +1008,16 @@ static void timer_stat_diff(
 				timer_stat_find(timer_stats_old, ts);
 			if (found) {
 				ts->delta = ts->count - found->count;
+				ts->time_delta = ts->time - found->time;
 				if (ts->delta >= opt_threshold) {
 					timer_stat_sort_freq_add(&sorted, ts);
 					sample_add(ts, whence);
 					found->info->total += ts->delta;
+					found->info->time_total += ts->time_delta;
 				}
 			} else {
 				ts->delta = ts->count;
+				ts->time_delta = duration;
 				if (ts->delta >= opt_threshold) {
 					timer_stat_sort_freq_add(&sorted, ts);
 					sample_add(ts, whence);
@@ -924,11 +1027,11 @@ static void timer_stat_diff(
 	}
 
 	if (!(opt_flags & OPT_QUIET)) {
-		unsigned long total = 0UL, kt_total = 0UL;
-		long int j = 0;
+		uint64_t total = 0UL, kt_total = 0UL;
+		int32_t j = 0;
 
 		printf("%8s %-5s %-15s",
-			opt_flags & OPT_CUMULATIVE ? "Events" : "Event/s", "PID", "Task");
+			(opt_flags & OPT_CUMULATIVE) ? "Events" : "Event/s", "PID", "Task");
 		if (!(opt_flags & OPT_BRIEF))
 			printf(" %-25s %-s\n",
 				"Init Function", "Callback");
@@ -939,9 +1042,9 @@ static void timer_stat_diff(
 			if (((n_lines == -1) || (j < n_lines)) && (sorted->delta != 0)) {
 				j++;
 				if (opt_flags & OPT_CUMULATIVE)
-					printf("%8lu ", sorted->count);
+					printf("%8" PRIu64, sorted->count);
 				else
-					printf("%8.2f ", (double)sorted->delta / dur);
+					printf("%8.2f ", (double)sorted->delta / duration);
 
 				if (opt_flags & OPT_BRIEF) {
 					char *cmd = sorted->info->cmdline ?
@@ -949,7 +1052,7 @@ static void timer_stat_diff(
 
 					printf("%5d %s\n",
 						sorted->info->pid,
-						opt_flags & OPT_CMD ?
+						(opt_flags & OPT_CMD) ?
 							cmd : sorted->info->task);
 				} else {
 					printf("%5d %-15s %-25s %-s\n",
@@ -963,10 +1066,11 @@ static void timer_stat_diff(
 
 			sorted = sorted->sorted_freq_next;
 		}
-		printf("%lu Total events, %5.2f events/sec (kernel: %5.2f, userspace: %5.2f)\n",
-			total, (double)total / dur,
-			(double)kt_total / dur,
-			(double)(total - kt_total) / dur);
+		printf("%" PRIu64 " Total events, %5.2f events/sec "
+			"(kernel: %5.2f, userspace: %5.2f)\n",
+			total, (double)total / duration,
+			(double)kt_total / duration,
+			(double)(total - kt_total) / duration);
 		if (!sane_procs)
 			printf("Note: this was run inside a container, kernel tasks were guessed.\n");
 		printf("\n");
@@ -978,20 +1082,22 @@ static void timer_stat_diff(
  *	scan /proc/timer_stats and populate a timer stat hash table with
  *	unique events
  */
-static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
+static void get_events(
+	timer_stat_t *timer_stats[],
+	const double time_now)
 {
 	FILE *fp;
 	char buf[4096];
 
-	if ((fp = fopen(TIMER_STATS, "r")) == NULL) {
-		fprintf(stderr, "Cannot open %s\n", TIMER_STATS);
+	if ((fp = fopen(proc_timer_stats, "r")) == NULL) {
+		fprintf(stderr, "Cannot open %s\n", proc_timer_stats);
 		return;
 	}
 
 	/* Originally from PowerTop, but majorly re-worked */
 	while (!feof(fp)) {
 		char *ptr = buf;
-		unsigned long count = -1;
+		uint64_t count = 0;
 		pid_t pid = -1;
 		char task[64];
 		char func[128];
@@ -1020,7 +1126,8 @@ static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 			continue;	/* Deferred event, skip */
 
 		ptr++;
-		sscanf(buf, "%21lu", &count);
+		if (sscanf(buf, "%21" SCNu64, &count) != 1)
+			continue;
 		memset(task, 0, sizeof(task));
 		memset(func, 0, sizeof(func));
 		memset(timer, 0, sizeof(timer));
@@ -1042,23 +1149,23 @@ static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
 			char tmp[64];
 			task[13] = '\0';
 			snprintf(tmp, sizeof(tmp), "[%s]", task);
-			strcpy(task, tmp);
+			strncpy(task, tmp, 13);
 		}
 
 		if (strcmp(task, "insmod") == 0)
-			strcpy(task, "[kern mod]");
+			strncpy(task, "[kern mod]", 13);
 		if (strcmp(task, "modprobe") == 0)
-			strcpy(task, "[kern mod]");
+			strncpy(task, "[kern mod]", 13);
 
 		if ((strncmp(func, "tick_nohz_", 10) == 0) ||
 		    (strncmp(func, "tick_setup_sched_timer", 20) == 0) ||
-		    (strcmp(task, APP_NAME) == 0))
+		    (strncmp(task, app_name, strlen(app_name)) == 0))
 			continue;
 
-		timer_stat_add(timer_stats, count, pid, task, func, timer, kernel_thread);
+		timer_stat_add(timer_stats, time_now, count, pid, task, func, timer, kernel_thread);
 	}
 
-	fclose(fp);
+	(void)fclose(fp);
 }
 
 /*
@@ -1067,8 +1174,8 @@ static void get_events(timer_stat_t *timer_stats[])	/* hash table to populate */
  */
 static void show_usage(void)
 {
-	printf("%s, version %s\n\n", APP_NAME, VERSION);
-	printf("Usage: %s [options] [duration] [count]\n", APP_NAME);
+	printf("%s, version %s\n\n", app_name, VERSION);
+	printf("Usage: %s [options] [duration] [count]\n", app_name);
 	printf("Options are:\n");
 	printf("  -c\t\treport cumulative events rather than events per second.\n");
 	printf("  -C\t\treport event count rather than event per second in CSV output.\n");
@@ -1086,11 +1193,12 @@ static void show_usage(void)
 int main(int argc, char **argv)
 {
 	timer_stat_t **timer_stats_old, **timer_stats_new, **tmp;
-	double duration_secs = 1.0;
-	long int count = 1;
-	long int n_lines = -1;
+	double duration_secs = 1.0, time_start, time_now;
+	int64_t count = 1, t = 1;
+	int32_t n_lines = -1;
 	bool forever = true;
-	struct timeval tv1, tv2, duration, whence;
+	struct sigaction new_action;
+	int i;
 
 	list_init(&timer_info_list);
 	list_init(&sample_list);
@@ -1117,7 +1225,7 @@ int main(int argc, char **argv)
 			eventstat_exit(EXIT_SUCCESS);
 		case 'n':
 			errno = 0;
-			n_lines = strtol(optarg, NULL, 10);
+			n_lines = (int32_t)strtol(optarg, NULL, 10);
 			if (errno) {
 				fprintf(stderr, "Invalid value for number of events to display\n");
 				eventstat_exit(EXIT_FAILURE);
@@ -1175,7 +1283,7 @@ int main(int argc, char **argv)
 	if (optind < argc) {
 		forever = false;
 		errno = 0;
-		count = strtol(argv[optind++], NULL, 10);
+		count = (int64_t)strtoll(argv[optind++], NULL, 10);
 		if (errno) {
 			fprintf(stderr, "Invalid count value\n");
 			eventstat_exit(EXIT_FAILURE);
@@ -1186,13 +1294,11 @@ int main(int argc, char **argv)
 		}
 	}
 
-	duration.tv_sec = (time_t)duration_secs;
-	duration.tv_usec = (suseconds_t)(duration_secs * 1000000.0) - (duration.tv_sec * 1000000);
 	opt_threshold *= duration_secs;
 
 	if (geteuid() != 0) {
 		fprintf(stderr, "%s requires root privileges to write to %s\n",
-			APP_NAME, TIMER_STATS);
+			app_name, proc_timer_stats);
 		eventstat_exit(EXIT_FAILURE);
 	}
 
@@ -1200,7 +1306,18 @@ int main(int argc, char **argv)
 	if (!sane_procs)
 		opt_flags &= ~(OPT_CMD_SHORT | OPT_CMD_LONG);
 
-	signal(SIGINT, &handle_sigint);
+	memset(&new_action, 0, sizeof(new_action));
+	for (i = 0; signals[i] != -1; i++) {
+		new_action.sa_handler = handle_sig;
+		sigemptyset(&new_action.sa_mask);
+		new_action.sa_flags = 0;
+
+		if (sigaction(signals[i], &new_action, NULL) < 0) {
+			fprintf(stderr, "sigaction failed: errno=%d (%s)\n",
+				errno, strerror(errno));
+			eventstat_exit(EXIT_FAILURE);
+		}
+	}
 
 	if ((timer_stats_old = calloc(TABLE_SIZE, sizeof(timer_stat_t*))) == NULL) {
 		fprintf(stderr, "Cannot allocate old timer stats table\n");
@@ -1212,54 +1329,55 @@ int main(int argc, char **argv)
 	}
 
 	/* Should really catch signals and set back to zero before we die */
-	set_timer_stat("1", true);
-	gettimeofday(&tv1, NULL);
-	get_events(timer_stats_old);
+	set_timer_stat("1\n", true);
+	time_now = time_start = gettime_to_double();
 
-	whence.tv_sec = 0;
-	whence.tv_usec = 0;
+	get_events(timer_stats_old, time_now);
 
 	while (!stop_eventstat && (forever || count--)) {
 		struct timeval tv;
+		double secs, duration = duration_secs;
 		int ret;
 
-		gettimeofday(&tv2, NULL);
-
-		tv = timeval_add(&duration, &whence);
-		tv = timeval_add(&tv, &tv1);
-		tv2 = tv = timeval_sub(&tv, &tv2);
-
+		/* Timeout to wait for in the future for this sample */
+		secs = time_start + ((double)t * duration_secs) - time_now;
 		/* Play catch-up, probably been asleep */
-		if (tv.tv_sec < 0) {
-			tv.tv_sec = 0;
-			tv.tv_usec = 0;
-			tv2 = tv;
+		if (secs < 0.0) {
+			t = ceil((time_now - time_start) / duration_secs);
+			secs = time_start + ((double)t * duration_secs) - time_now;
+			/* Really, it's impossible, but just in case.. */
+			if (secs < 0.0)
+				secs = 0.0;
+		} else {
+			t++;
 		}
-
-		ret = select(0, NULL, NULL, NULL, &tv2);
+		tv = double_to_timeval(secs);
+		ret = select(0, NULL, NULL, NULL, &tv);
 		if (ret < 0) {
 			if (errno == EINTR) {
-				duration = timeval_sub(&tv, &tv2);
-				stop_eventstat = true;
+				goto abort;
 			} else {
-				fprintf(stderr, "Select failed: %s\n", strerror(errno));
+				fprintf(stderr, "select() failed: errno=%d (%s)\n",
+					errno, strerror(errno));
 				break;
 			}
 		}
 
-		get_events(timer_stats_new);
-		timer_stat_diff(&duration, n_lines, &whence,
+		duration = gettime_to_double() - time_now;
+		duration = floor((duration * 100.0) + 0.5) / 100.0;
+		time_now = gettime_to_double();
+
+		get_events(timer_stats_new, time_now);
+		timer_stat_diff(duration, n_lines, time_now,
 			timer_stats_old, timer_stats_new);
 		timer_stat_free_contents(timer_stats_old);
 
 		tmp             = timer_stats_old;
 		timer_stats_old = timer_stats_new;
 		timer_stats_new = tmp;
-
-		whence = timeval_add(&duration, &whence);
 	}
-
-	samples_dump(csv_results, &duration);
+abort:
+	samples_dump(csv_results);
 
 	timer_stat_free_contents(timer_stats_old);
 	timer_stat_free_contents(timer_stats_new);
